/*
 * audio.c for miniD pre OLED
 *
 * Audio related functionality: control SRC (sample rate converter), DAC (DA-converter),
 * and PGA, calculate volume values, mute control
 *
 * Initial version: 20/12/2017
 * Author: Tomi Nihtila, nihtila.com
 */ 

#include "audio.h"
#include "globals.h"
#include "spi.h"
#include "main.h"
#include <avr/io.h>
#include <stdbool.h>
#include <util/delay.h>

// Chip-level communication is kept internal
static void audioSRCWrite(unsigned char address, unsigned char data);
static unsigned char audioSRCRead(unsigned char address);
static void audioDACWrite(unsigned char address, unsigned char data);
static void audioUpdateDigitalVol(signed char *vol);
static void audioUpdateAnalogVol(signed char *vol);



// ********************************************************
// INITIALISE AUDIO
// Initialise SPI and then DAC and SRC chips
//
void audioInit(void) {
	
	signed char minVol[] = {-120, -120, -120, -120};
	
  
  // Initialise SPI block
	SPIInit();
	
  
	// DAC INIT
  
	// If digital volume control is selected, it is possible there is no PGA4311 chip
	// populated, meaning volume would be at maximum in the beginning. Set minimum
	// volume on DAC before configuring the interface.
	if(sys.menuItem[VOLCONT_TYPE_MI].data == VOLCONT_DIGITAL) {
		audioUpdateDigitalVol(minVol);
	}
	// Change data format to I2S, everything else in DAC is ok by default.
	audioDACWrite(0x07, 0x01);
	
  
	// SRC INIT
	// SRC takes in one of four SPDIF inputs and converts sample rate to 48k required
	// by the miniDSP. Routing is DIR-SRC-PORTA. PORTB is unused. MCLK is external and 
	// PORTA is I2S master.
  
	// Software reset
	audioSRCWrite(0x01, 0x80);
	// Register 03; Port A: format I2S, master mode, src->porta
	audioSRCWrite(0x03, 0x39);
	// Register 04; port A clock to MCLK/256 (48k)
	audioSRCWrite(0x04, 0x01);
	// Register 05: Port B: format I2S, slave mode, mute output (as port is only input)
	audioSRCWrite(0x05, 0x40);
	// Register 0D; DIR reference clock = MCLK
	audioSRCWrite(0x0D, 0x08);
	// Register 0E; automatic DIR-mute for loss of lock
	audioSRCWrite(0x0E, 0x08);
	// Registers 0F, 10 and 11; PLL1 configuration. Values from datasheet: P=1, J=8, D=0
	audioSRCWrite(0x0F, 0x12);
	audioSRCWrite(0x10, 0);
	audioSRCWrite(0x11, 0);
	// Register 16 and 17; unmask receiver interrupts - the idea is to signal DIR unlock
	// and some real errors
	audioSRCWrite(0x16, AUDIO_SRC_R14_INTMASK);
	// Registers 18-1A; interrupt mode to level active
	audioSRCWrite(0x18, AUDIO_SRC_R18_INTMODE);
	audioSRCWrite(0x19, AUDIO_SRC_R19_INTMODE);
	// Register 2D; dir->src
	audioSRCWrite(0x2D, 0x02);
	
	// Register 01; power-up SRC, DIR, and port A (and B)
	//audioSRCWrite(0x01, 0x33);
	audioSRCWrite(0x01, 0x3B);
}


// ********************************************************
// UPDATE AUDIO INPUT
// Changes input if it has been changed in program. Also checks that disabled input is
// not selected (this function is called every time input is changed).
//
void audioUpdateInput(void) {
	
	unsigned char temp;
	static char oldInput = 100;		// initialise with invalid value
	
	
	// First ensure disabled input is not selected - 0=OP1, 1=OP2, 2=CX1, 3=CX2, 4=ANA
	if(preamp.input == 1 && sys.menuItem[USE_OP2_MI].data == 0) {
		preamp.input++;
	}
	if(preamp.input == 3 && sys.menuItem[USE_CX2_MI].data == 0) {
		preamp.input++;
	}
	if(preamp.input == 4 && sys.menuItem[USE_ANA_MI].data == 0) {
		preamp.input++;
	}
	if(preamp.input >= preamp.noInputs) {
		preamp.input = 0;	
	}
	
	
	// If input has changed since previous update, change it
	if(preamp.input != oldInput) {
	
		// Analog input
		if(preamp.input == 4) {
			//PORTD |= 0x04;		// DSP_SEL2	selects analog - does not work!
			
			// Changing to analog input requires re-routing in SRC: ADC signal from
			// miniDSP goes to Port B of SRC. Therefore, source of Port A must be
			// changed from SRC to Port B (no sample-rate conversion needed).
			
			// Change Port A source from SRC to Port B (format still I2S and master mode)
			audioSRCWrite(0x03, 0x19);
			
			// unlock etc are not relevant with analog input
			resetEvent(SRC_INT_EV, false, false);			
			preamp.signalLock = true;
			sys.dispBlink = false;
			sys.errorCode = 0x00;
		}
		// Digital input
		else {
			//PORTD &= ~0x04;		// DSP_SEL2 selects digital

			// Change Port A source from SRC to Port B (format still I2S and master mode)
			audioSRCWrite(0x03, 0x39);
			
			// Also call SRC interrupt event as it was disabled for analog input
			resetEvent(SRC_INT_EV, true, false);
		
			// Select specific digital input with SRC register
			// Value needs some mapping since on PCB 0=CX1, 1=OP1, 2=OP2, 3=CX2
			// but in software 0=OP1, 1=OP2, 2=CX1, 3=CX2
			if(preamp.input == 0) {
				temp = 1;
			}
			else if(preamp.input == 1) {
				temp = 2;
			}
			else if(preamp.input == 2) {
				temp = 0;
			}
			else {
				temp = 3;
			}
			
			// Register 0D is input selection
			temp |= 0x08;		  // select MCLK for DIR (not related to input selection)
			audioSRCWrite(0x0D, temp);
		}
		
		oldInput = preamp.input;
    
		// If input changed, different offset may require changing volume
		audioUpdateVolume();    
	}

}


// ********************************************************
// UPDATE AUDIO VOLUME
// This is called every time volume, offset, or balance is changed. Therefore, it checks master 
// volume limits and computes volume for each channel. If they have changed since the last call,
// send new values forward to be changed in analog or digital domain.
//
void audioUpdateVolume(void) {
	
	static signed char oldVol[] = {MAX_OFFSET+1, MAX_OFFSET+1, MAX_OFFSET+1, MAX_OFFSET+1}; // invalid init
	signed char vol[4], defaultVol, balance;
	unsigned char i;
	bool update;
	
	
	// CHECK LIMITS
	// Update maximum volume limit
	preamp.maxVol = sys.menuItem[VOL_LIMIT_MI].data;
	
	// Ensure current master volume is within limits (minVol-1 = mute)
	if(preamp.masterVol <= preamp.minVol-1) {
		preamp.masterVol = preamp.minVol-1;
		preamp.mute = true;
		audioMuteControl();			// mute
	}
	else if(preamp.masterVol > preamp.maxVol) {
		preamp.masterVol = preamp.maxVol;
	}
	else {	// volume within range
		preamp.mute = false;
		// only call unmute if it was just changed (mute signal is low)
		if((PORTD & 0x80) == 0) {
			audioMuteControl();		// unmute
		}
	}
	
	
	// COMPUTE VOLUME FOR EACH CHANNEL
	// This takes offsets and balance into account. Note that with this implementation
	// individual channel can get higher gain value than set maximum volume.
	
	// Basis of every setting: master volume + offset of the selected input
	// (this line assumes OP1 is always the first input)
	defaultVol = preamp.masterVol + sys.menuItem[OP1_OS_MI+preamp.input].data;
	
	// Channels 1&2 (L & R)
	// Note reversed order.
	
	// This is done as balance used to be -x...x but is now 0...2x
	balance = sys.menuItem[OUT1_BAL_MI].data - MAX_BALANCE;
	
	// If balance is positive, left channel is attenuated (0: nothing changes)
	if(balance >= 0)
	{
		// Balance is positive, meaning right is louder -> attenuate left (1)
		vol[1] = defaultVol - balance;
		// Right (2) channel
		vol[3] = defaultVol;
	}
	// If balance is negative, right channel is attenuated
	else
	{
		// Left (1) channel
		vol[1] = defaultVol;
		// Right (2); balance sign is now reversed since its negative
		vol[3] = defaultVol + balance;
	}
		
	// Channels 3&4 (L & R)
	// Basis is previous one + output2 offset
	defaultVol = defaultVol + sys.menuItem[OUT2_OS_MI].data;
		
	// Balance goes in similar way than before
	balance = sys.menuItem[OUT2_BAL_MI].data - MAX_BALANCE;
	
	if(balance >= 0)
	{
		vol[2] = defaultVol - balance;
		vol[0] = defaultVol;
	}
	else
	{
		vol[2] = defaultVol;
		vol[0] = defaultVol + balance;
	}
			

	// If any value has changed, update all volume settings
	update = false;
	for(i = 0; i < 4; i++) {
		// When this is commented out, gain in one channel may be over MAX_GAIN if there is positive
		// offset. However, if all channels are forced within MAX_GAIN, balance and offset proportions
		// "break" near MAX_GAIN.
		/*
		if(vol[i] < MIN_GAIN) {
			vol[i] = MIN_GAIN;
		}
		else if(vol[i] > MAX_OFFSET) {
			vol[i] = MAX_GAIN;
		}
		*/
		if(vol[i] != oldVol[i]) {
			update = true;
		}
		
		oldVol[i] = vol[i];
	}
	
	if(update == true) {
		// Something has been changed, call actual volume update. Function depends if using analog or
		// digital volume control.
		if(sys.menuItem[VOLCONT_TYPE_MI].data == VOLCONT_DIGITAL) {
			audioUpdateDigitalVol(vol);
		}
		else {
			audioUpdateAnalogVol(vol);
		}
	}
}


// ********************************************************
// UPDATE DIGITAL VOLUME
// Write attenuation values in DAC.
//
static void audioUpdateDigitalVol(signed char *vol) {
	
	unsigned char i, temp;

	
	for(i = 0; i < 4; i++) {
		// Map volume values into PCM4101-compatible values
		temp = (signed int)vol[i]*2+255;
    
		// Attenuation registers are in addresses 1-4
		audioDACWrite(i+1, (unsigned char)temp);
	}

}


// ********************************************************
// UPDATE ANALOG VOLUME
// Write attenuation values in PGA4311.
//
static void audioUpdateAnalogVol(signed char *vol) {
	
	unsigned char i, temp;
	
	
	// There is no separate write function for PGA4311 as this is the only place it is needed
	// and it writes 4 consecutive bytes.
	SPIStart(SPI_CS_PGA);
		
	for(i = 0; i < 4; i++) {
		// Map volume values into PGA4311-compatible values
		temp = (vol[i]+96)*2;
    
		SPITransferByte(temp);
	}
		
	SPIEnd(SPI_CS_PGA);	
}


// ********************************************************
// CHANGE VOLUME CONTROL TYPE
// Called when switching between analog and digital volume control. This sets the gain 
// of non-active control to 0dB and calls volume update.
//
void audioChangeVolumeControl(void) {
	
	signed char vol[] = {0, 0, 0, 0};
	
	
	// Call to update the active control. The routine does not update the
	// volume if values have not been changed; therefore there is change of volume and
	// double update.
	preamp.masterVol--;
	audioUpdateVolume();
	preamp.masterVol++;
	audioUpdateVolume();
	
	// When changing from analog to digital control, digital attenuation is slowly changed
	// so it is still almost at maximum level when analog is returned to 0dB, causing
	// high level signal for a short period of time. This delay helps in that. It is a
	// dummy delay but it's ok here, this is not used often anyway.
	_delay_ms(5);
	
	// Set the inactive control to maximum
	if(sys.menuItem[VOLCONT_TYPE_MI].data == VOLCONT_ANALOG) {
		// Analog control chosen, set digital gain to 0dB
		audioUpdateDigitalVol(vol);
	}
	else {
		// Digital chosen, set analog gain to 0dB (if the chip is even populated)
		audioUpdateAnalogVol(vol);
	}	
}


// ********************************************************
// MUTE CONTROL
// Mute/unmute all channels or output2 depending on the controls. Controls mute in both
// PGA and DAC.
// 
void audioMuteControl(void) {
	
	// If mute off
	if(preamp.mute == false) {

		// Check if output 2 should be muted (in DAC)
		if(sys.menuItem[OUT2_MUTE_MI].data == 1) {
			audioDACWrite(0x05, 0xC0);	// 3&4 muted
		}
		else {
			audioDACWrite(0x05, 0x00);	// all on
		}		
		
		// Unmute PGA
		PORTD |= 0x80;		
	}
	else {
		// Mute PGA
		PORTD &= ~0x80;
		
		// Mute all DAC channels (so that mute works also if PGA is not used)
		if(sys.menuItem[VOLCONT_TYPE_MI].data == VOLCONT_DIGITAL) {
			audioDACWrite(0x05, 0xF0);
		}		
	}
	
}


// ********************************************************
// DSP BYPASS CONTROL
// Controls DSP bypass switch.
//
void audioDSPBypassControl(void) {
	
	// DSP_SEL1 controls DSP bypass
	if(sys.menuItem[DSP_BYP_MI].data == 1) {
		PORTD |= 0x02;
	}
	else {
		PORTD &= ~0x02;		
	}
	
}


// ********************************************************
// UPDATE SRC SETTINGS
// Updates de-emphasis filter and group delay filter settings in SRC.
//
void audioSRCChangeSettings(void) {
	
	unsigned char temp;
	
	// De-emphasis filter
	temp = sys.menuItem[DE_EMPH_MI].data << 3;
	
	// Filter group delay
	temp |= sys.menuItem[SRC_FGD_MI].data;
	
	audioSRCWrite(0x2E, temp);
}


// ********************************************************
// SRC WRITE
// Writes byte to SRC address.
//
static void audioSRCWrite(unsigned char address, unsigned char data) {
	
	SPIStart(SPI_CS_SRC);
	
	SPITransferByte(address);
	SPITransferByte(0);			// SRC has don't care byte for backward compatibility
	SPITransferByte(data);
	
	SPIEnd(SPI_CS_SRC);
	
}


// ********************************************************
// SRC READ
// Reads byte from SRC address.
//
static unsigned char audioSRCRead(unsigned char address) {
	
	unsigned char data;

  
	SPIStart(SPI_CS_SRC);
	
	address |= 0x80;			// set read bit
	SPITransferByte(address);
	SPITransferByte(0);			// SRC has don't care byte for backward compatibility
	data = SPITransferByte(0);	// dummy write for reading data
	
	SPIEnd(SPI_CS_SRC);
	
	return data;
}


// ********************************************************
// DAC WRITE
// Writes byte to DAC address.
//
static void audioDACWrite(unsigned char address, unsigned char data) {
	
	address |= 0x20;			// Requires extra bit set
	
	SPIStart(SPI_CS_DAC);
	
	SPITransferByte(address);
	SPITransferByte(data);
	
	SPIEnd(SPI_CS_DAC);
	
}


// ********************************************************
// SRC INTERRUPT HANDLER
// Called when external INT line is active. Checks the source of interrupt and
// raises unlock flag or error condition.
//
void audioSRCInt(void) {
	
	unsigned char r14;
	
	
	// When SRC interrupt is set level active, first it becomes active when an interrupt flag is set.
	// When R14 is read, it clears the interrupts but if the condition still exist, the flag is re-set
	// immediately. As this function is called by pin change interrupt, it causes constant re-call 
	// unless simple manual flag check and clear is performed.
	if((PINB & 0x01) == 0) {

		// Reading R14 clears interrupt flags and if the condition still exists, they are re-set immediately
		// causing another PCINT interrupt on INT line. The programming logic must be so that no odd nested
		// calls can occur (as I had a combination of normal calls and calls in ISR to this function, using
		// audioSRCInt as event function prevents this).
		r14 = audioSRCRead(0x14);		// DIR error status register
		r14 &= AUDIO_SRC_R14_INTMASK;   // mask some bits, 1 indicates an error
	
		if(r14!= 0) {	
			// An error bit is set - check if it is DIR unlock
			if((r14 & 0x04) != 0) {
				// This causes input symbol to blink on display
				preamp.signalLock = false;
				
				// display update needs to be called the first time to start blinking (only in basic state)
				if((sys.events[DISP_REFRESH_EV].execute == false) && (sys.state == VOLUME_ST)) {
					resetEvent(DISP_REFRESH_EV, true, true);	
				}
			}
			else {	// real error other than unlock

				sys.state = ERROR_ST;
				sys.errorCode = r14;
			}
		}
		resetEvent(SRC_INT_EV, true, false);
	}
	else { // High = signal is locked and no errors, back to normal state
		preamp.signalLock = true;
		sys.dispBlink = false;
		sys.errorCode = 0x00;
		// If in error state, return to normal. But do not return if in menu for example.
		if(sys.state == ERROR_ST) {
			sys.state = VOLUME_ST;
		}
		
		resetEvent(SRC_INT_EV, false, false);
		// error state is over - update display
		resetEvent(DISP_REFRESH_EV, true, true);	
		
		audioGetSignalInfo();
	}	
	
}


// ********************************************************
// GET INPUT SIGNAL INFO
// Calculates input signal sample rate from SRC status information.
//
void audioGetSignalInfo(void) {
	
	unsigned char intRatio;
	unsigned int fracRatio;
	unsigned long int fs;
	
	
	// R32 bits 7-3 are integer part, bits 2-0 and R33 are the fractional part
	intRatio = audioSRCRead(0x32);
	
	fracRatio = (intRatio & 0x07) << 8;
	fracRatio |= (unsigned int)audioSRCRead(0x33);
	
	intRatio >>= 3;
	
	fs = ((unsigned long int)fracRatio*48)/2047;	// 11-bit fractional part is fracRatio/2027 * 48k
	fs += intRatio*48;
	
	preamp.inputFs = (unsigned char)fs;
}