/*
 * eeprom.c for miniD pre OLED
 *
 * Data storing to and retrieving from EEPROM.
 *
 * Initial version: 20/12/2017
 * Author: Tomi Nihtila, nihtila.com
 */ 

#include "eeprom.h"
#include "globals.h"
#include <avr/eeprom.h>
#include <stdbool.h>


// ********************************************************
// SAVE SETTINGS TO EEPROM
// Stores all menu settings and current input. Ensure these are read in the same 
// order in readEEPROM()!
//
void saveToEEPROM(void) {

	unsigned char* address = 0;
	unsigned char i;
	
	// Input is also used as a flag when checking if EEPROM is written at startup.
	// Update method where data is written only if it has changed is used.
	eeprom_update_byte(address, preamp.input);
	address++;                                  // input variable size is one byte
	
	// Store all menuitem data fields
	for(i = 0; i < sys.noMenuItems; i++) {
		eeprom_update_byte(address, sys.menuItem[i].data);
		address++;                              // data size is also one byte
	}

	// Store IR codes which are a bit of an exception to the above
	for(i = 0; i < sys.ir.noCodes; i++) {
		// dword is double word = 4 bytes
		eeprom_update_dword((unsigned long int*)address, sys.ir.savedCodes[i]);
		address += 4;                           // one code is 4 bytes
	}

}


// ********************************************************
// READ SETTINGS FROM EEPROM
// Reads settings and current input from EEPROM. Ensure these are written in the
// same order in saveToEEPROM()! Returns true if settings were found in EEPROM, 
// false if not.
//
bool readEEPROM(void) {

	unsigned char* address = 0;
	unsigned char firstbyte;
	unsigned char i;
	
  
	// Input is also used as a flag when checking if EEPROM is written at startup
	firstbyte = eeprom_read_byte(address);

	// Unwritten EEPROM reads 0xFF (input can never be 0xFF), meaning no values have been saved
	if(firstbyte != 0xFF) {
		preamp.input = firstbyte;

		address++;                                  // input variable size is one byte
		
		// Read all menuitem data fields
		for(i = 0; i < sys.noMenuItems; i++) {
			sys.menuItem[i].data = (signed char)eeprom_read_byte(address);
			address++;                              // one byte data
		}

		// read IR codes
		for(i = 0; i < sys.ir.noCodes; i++) {
			sys.ir.savedCodes[i] = eeprom_read_dword((unsigned long int*)address);
			address += 4;                           // one code is 4 bytes
		}
		
		return true;
	}
	else
		return false;

}
