/*
 * globals.h for miniD pre OLED
 *
 * Global defines and typedefs for global structs.
 *
 * Initial version: 03/12/2017
 * Author: Tomi Nihtila, nihtila.com
 */ 


#ifndef GLOBALS_H_
#define GLOBALS_H_

#include <stdbool.h>
#include <avr/pgmspace.h>

#define F_CPU	8000000UL

// These are defined by the hardware. MAX_GAIN comes from the maximum gain set by resistors and minimum gain is
// steps*stepsize smaller. Default max gain is 6 dB and min gain -57 dB (64 steps, 1 dB each).
#define MAX_GAIN	10
#define MIN_GAIN	-80	// One step smaller to this activates mute; hardware limit for this is -95
#define DEF_GAIN	-40

#define MIN_OFFSET	-10
#define MAX_OFFSET	10

#define MAX_BALANCE	5		// Note that changing this requires changing how display shows balance

#define NO_INPUTS	5

// One timer tick in ms. This does not control timer setup - change registers if you want to change the tick value.
#define TICK      2



typedef enum {
	LOW = 0,
	HIGH = 1
} ioState;


// SYSTEM STATE NAMES/INDICES
typedef enum {
	STARTUP_ST,
	VOLUME_ST,                          // Basic state when display is showing volume and input
	MENU_ST,                            // In menu, selecting menu item to change
	MENUITEM_ST,                        // When menu item is selected and changing the value
	ERROR_ST							// When error occurs in SRC (other than DIR unlock)
} sysState;


// VOLUME CONTROL TYPE
typedef enum {
	VOLCONT_ANALOG = 0,
	VOLCONT_DIGITAL = 1
} volContType;


// MENU ITEM NAMES/INDICES
// NOTE THAT THIS ORDER MUST MATCH THE ORDER OF SYS INIT IN MAIN.C!
typedef enum {							// Enum names for the menu items, used to control what is on display
	DSP_BYP_MI,
	OUT2_MUTE_MI,
	OUT2_OS_MI,
	VOLCONT_TYPE_MI,
	USE_OP2_MI,
	USE_CX2_MI,
	USE_ANA_MI,
	STARTUP_VOL_MI,
	VOL_LIMIT_MI,
	OP1_OS_MI,
	OP2_OS_MI,
	CX1_OS_MI,
	CX2_OS_MI,
	ANA_OS_MI,
	OUT1_BAL_MI,
	OUT2_BAL_MI,
	STARTUP_DEL_MI,
	SRC_FGD_MI,
	DE_EMPH_MI,
	LEARN_IR_MI
} menuItem;
#define NO_MENUITEMS  LEARN_IR_MI+1		// Number of menu items - assumes last item wont be changed!


// EVENT NAMES/INDICES
typedef enum {          // Enum names for events
	STARTUP_EV,         // 0	startup
	ENCODER_CHANGE_EV,	// 1	encoder position change
	ENCODER_TIMEOUT_EV, // 2	encoder timeout
	SH_BUTTON_EV,		// 3	encoder short button push (long is handled by button push counter)
	BUTTON_PUSH_EV,     // 4	encoder button push counter
	IR_ACTION_EV,		// 5	interpret IR code
	IR_TIMEOUT_EV,      // 6	IR timeout
	SRC_INT_EV,			// 7	SRC interrupt check
	DISP_REFRESH_EV		// 8	display update
} eventNames;
#define NO_EVENTS DISP_REFRESH_EV+1     // Number of different events - keep disp update last!


typedef enum {							// IR commands that can be taught and used
	UP_IR,
	DOWN_IR,
	SH_BUT_IR,
	LN_BUT_IR
} irCommands;
#define NO_IR_CMD 4						// Number of IR commands - change if adding items!


// MENU ITEM STRUCT
typedef struct {						// Variables for each menu item, used in system struct
	const char* const text;				// menu text (pointer to string in PROGMEM)
	signed char data;                   // actual data
	signed char dataMin;                // data minimum value
	signed char dataMax;                // data maximum value
	const char* const unit;				// unit (pointer to string in PROGMEM)
	const char* const* dataMap;			// strings showed instead of numerical data 
										//   (pointer to array of pointers (in PROGMEM) to strings in PROGMEM)
	const unsigned char* const xLoc;	// X-location for data/unit in display (pointer to uint8 in PROGMEM)
	void (*dataFunc)();                 // pointer to function which is called when data is changed	
} menuItemSt;


// EVENT STRUCT
typedef struct {                        // Variables for each event, used in system struct
	unsigned int ticks;                 // time (in x ms ticks) when event function is executed if execute is set to true
	bool execute;
	const unsigned int interval;        // if execute is left true, event is set to occur at fixed time intervals (in x ms ticks)
	void (*func)();                     // pointer to function which is called when event is executed
} eventSt;


// IR STRUCT
typedef struct {						// IR struct, used in system struct
	bool repeatCode;					// repeat code yes/no
	bool timeout;						// timeout signals something in the reception went wrong
	const unsigned char noCodes;		// number of IR codes
	unsigned long int code;				// 32-bit IR code that was just received
	unsigned long int savedCodes[NO_IR_CMD];	// saved codes for the available commands
} irSt;



// SYSTEM
// Keeps track of system state and holds simple scheduler that calls events to execute at set time intervals. Uses the structs
// and enums defined above. Initialised in main.c
typedef struct {
	sysState state;                     // system state
	volatile unsigned int ticks;		// free-running counter, time base for the scheduler
	bool dispBlink;						// controls display blinking when required
	unsigned char menuIndex;            // index for menu items when in menu
	const unsigned char noMenuItems;    // number of menu items
	menuItemSt menuItem[NO_MENUITEMS];  // holds data structure for each menu item, defined above
	eventSt events[NO_EVENTS];          // holds data structure for each event, defined above
	irSt ir;                            // IR remote data structure, defined above
	unsigned char errorCode;			// SRC error code
} sysSt;


// PREAMPLIFIER
// Keeps track of audio related status and data. Initialised in main.c
typedef struct {
	signed char masterVol;              // current master volume (ignoring offsets, balance, etc.)
	signed char minVol;                 // minimum volume
	signed char maxVol;                 // maximum volume, limit can be set in menu
	unsigned char input;                // current input
	bool mute;							// mute on/off
	const unsigned char noInputs;       // number of inputs
	const char *inputName[5];			// names for inputs (could be in PROGMEM but is quite small..)
	bool signalLock;					// SPDIF signal lock flag
	unsigned char inputFs;				// input fs (also used as flag, hold value only momentarily)
} preampSt;


// These are used globally in all code blocks
extern sysSt sys;
extern preampSt preamp;


// This is related to system event timing but is also used in encoder.c
#define ENC_BUT_CHECK_INTERVAL	10/TICK

#endif /* GLOBALS_H_ */