/*
 * ir.c for miniD pre OLED
 *
 * IR remote control for NEC IR codes.
 *
 * Initial version: 03/12/2017
 * Author: Tomi Nihtila, nihtila.com
 */ 

#include "ir.h"
#include "main.h"
#include "globals.h"
#include "encoder.h"
#include <avr/io.h>
#include <stdbool.h>


// State of IR receiver
typedef enum {
	WAIT_IR,
	DECODE_IR
} irState;



// ********************************************************
// IR RECEIVER (ISR)
// Does simple calculation of duration of low/high state and decodes that into bits. This is 
// called inside ISR so it must be relatively fast. This implementation decodes NEC IR code.
//
void IRISR(void) {

	static unsigned char lastCnt = 0;
	static unsigned int lastTick = 0;
	static unsigned long int data;
	static unsigned char bitCnt;
	static irState irReceiver = WAIT_IR;
	unsigned int dTime;


	if(sys.ir.timeout == false) {
		// Calculate time in us ticks since the last ISR call. Note that the resolution is only one step
		// of timer counter which by default is 128 us.
		// Note: in BHT preamp it was 64 us, can we overflow or miss something?
		//        (dTimer            dTicks * timer compare) * us / timer step
		dTime = (TCNT2-lastCnt + (sys.ticks-lastTick)*OCR2A) * (TICK*1000/OCR2A);
	}
	// If timeout has occurred, dTime would overflow. Therefore, set dTime to max value.
	else {
		dTime = 0xFFFF;
		sys.ir.timeout = false;
	}

	lastCnt = TCNT2;
	lastTick = sys.ticks;


	// If waiting for data decoding
	if(irReceiver == WAIT_IR) {
	
		// Check the length of the space after start pulse. For simplicity, 9ms start pulse
		// is not checked at all.
		if((PIND & 0x40) == 0) {

			// In case of valid 4500 us space, initialise variables and wait for data decoding.
			// Note: limits were widened as miniD uses inaccurate internal RC oscillator.
			//if(dTime > 4000 && dTime < 5000) {
			if(dTime > 3500 && dTime < 5500) {
				irReceiver = DECODE_IR;
				data = 0;
				bitCnt = 0;
			}
			// Repeat pulse
			//else if(dTime > 2000 && dTime < 2500 && sys.ir.code != 0) {
			else if(dTime > 1700 && dTime < 2800 && sys.ir.code != 0) {
				sys.ir.repeatCode = true;
				resetEvent(IR_ACTION_EV, true, true);	// code ready
			}
			else {
				irReceiver = WAIT_IR;
			}
		}
	}
	// Receiving data
	else {
	
		// 562.5 us pulses are not checked at all but the space after which defines logical 0 or 1.
		// Space in 0 is 562.5 us and in one 1687.5 us.
		if((PIND & 0x40) == 0) {

			// If time between pulses have been long, something has gone wrong
			if(dTime > 50000) {
				irReceiver = WAIT_IR;
			}
			// Total stream consists of 8-bit address and its inverse, and 8-bit command and its inverse
			else {
				// Logical 1
				if(dTime > 1100) {
					data |= 0x01;
				}
				
				if(bitCnt < 31) {
					data <<= 1;
				}
				else {
					// Set new code and signal that it is ready
					sys.ir.code = data;
					// move to wait for next transmission
					irReceiver = WAIT_IR;
					resetEvent(IR_TIMEOUT_EV, false, false);
					resetEvent(IR_ACTION_EV, true, true);
				}
				bitCnt++;
			}
		}
	}

	// Set IR timeout event that clears IR variables after a small delay in case
	// reception is left in a weird unfinished state. In normal reception this is
	// called often enough that timeout never occurs. Note that timeout must be set
	// longer than delay between repeat pulses.
	resetEvent(IR_TIMEOUT_EV, true, false);
}


// ********************************************************
// IR TIMEOUT
// Resets IR variables. Part of the timeout recovery is in IRISR().
//
void irTimeout(void) {
	
	// Reset IR variables
	sys.ir.code = 0;
	sys.ir.timeout = true;
	
	resetEvent(IR_TIMEOUT_EV, false, false);
}



// ********************************************************
// IR ACTION
// This is called when code is ready. It can be used for code learning or
// performing an action depending on the system state.
//
void irAction(void) {

	// If in menu learning IR code
	if((sys.state == MENUITEM_ST) && (sys.menuIndex == LEARN_IR_MI)) {

		if(sys.ir.repeatCode == false) {
		  
			// Store code
			sys.ir.savedCodes[sys.menuItem[LEARN_IR_MI].data] = sys.ir.code;
			
			// Signal user code was successfully received by moving to next command, 
			// or back to menu if it was the last one
			if(sys.menuItem[LEARN_IR_MI].data < sys.menuItem[LEARN_IR_MI].dataMax) {
				sys.menuItem[LEARN_IR_MI].data++;
			}
			else {
				sys.menuItem[LEARN_IR_MI].data = 0;
				sys.state = MENU_ST;
			}

			sys.dispBlink = false;
			resetEvent(DISP_REFRESH_EV, true, true);
		}
	}
	// If not in learning mode, perform action if code is found in saved codes list
	else {
		if(sys.ir.code == sys.ir.savedCodes[UP_IR]) {
			if(sys.state == MENU_ST) {
				encoder.counter -= 4;
			}
			else {
				encoder.counter += 4;
			}
			encoderChange();
		}
		else if(sys.ir.code == sys.ir.savedCodes[DOWN_IR]) {
			if(sys.state == MENU_ST) {
				encoder.counter += 4;
			}
			else {
				encoder.counter -= 4;
			}
			encoderChange();
		}
		else if(sys.ir.code == sys.ir.savedCodes[SH_BUT_IR] && sys.ir.repeatCode == false) {
			encoderShortButtonPush();
		}
		else if(sys.ir.code == sys.ir.savedCodes[LN_BUT_IR] && sys.ir.repeatCode == false) {
			encoderLongButtonPush();
		}

	}

	sys.ir.repeatCode = false;
	resetEvent(IR_ACTION_EV, false, false);
}