/*
 * oled.c for miniD pre OLED
 *
 * OLED display controls using I2C OLED and u8g2 library.
 *
 * Initial version: 20/12/2017
 * Author: Tomi Nihtila, nihtila.com
 */ 

#include "oled.h"
#include "globals.h"
#include "u8g2/u8g2.h"
#include "main.h"
#include <avr/io.h>
#define __DELAY_BACKWARD_COMPATIBLE__
#include <util/delay.h>
#include <stdio.h>
#include <stdbool.h>
#include <avr/pgmspace.h>


#define OLED_ROW_BUF	12

static void oledUpdateVolIn(void);
static void oledUpdateMenu(void);
static void oledUpdateMenuItem(void);
static void oledUpdateError(void);


// Functions related to u8g2 that needed to be implemented outside the library
uint8_t u8g2_gpio_delay_sw_i2c(u8x8_t *u8x8, uint8_t msg, uint8_t arg_int, void *arg_ptr);
uint8_t u8g2_gpio_delay_dummy(u8x8_t *u8x8, uint8_t msg, uint8_t arg_int, void *arg_ptr);
uint8_t u8g2_avr_hw_i2c(u8x8_t *u8x8, uint8_t msg, uint8_t arg_int, void *arg_ptr);

static u8g2_t disp;   // display object



// ********************************************************
// INITIALISE OLED
//
void oledInit(void) {
	
	u8g2_Setup_ssd1306_i2c_128x64_noname_f(&disp, U8G2_R0, u8g2_avr_hw_i2c, u8g2_gpio_delay_dummy);
	//u8g2_Setup_sh1106_i2c_128x64_noname_f(&disp, U8G2_R0, u8g2_avr_hw_i2c, u8g2_gpio_delay_dummy);
	//u8g2_Setup_ssd1306_i2c_128x64_noname_f(&disp, U8G2_R0, u8x8_byte_sw_i2c, u8g2_gpio_delay_sw_i2c);
	u8g2_InitDisplay(&disp);
	u8g2_SetPowerSave(&disp, 0);	
}


// ********************************************************
// OLED UPDATE
// Call display update function based on current system state.
//
void oledUpdate(void) {
	
	u8g2_ClearBuffer(&disp);

	// VOLUME AND INPUT (basic state)
	if(sys.state == VOLUME_ST || sys.state == STARTUP_ST) {

		oledUpdateVolIn();
	}
	// MENU
	else if(sys.state == MENU_ST) {
		
		oledUpdateMenu();
	}
	// MENU ITEM
	else if(sys.state == MENUITEM_ST) {

		oledUpdateMenuItem();
	}
	// ERROR
	else if(sys.state == ERROR_ST) {
		
		oledUpdateError();
	}
	
	u8g2_SendBuffer(&disp);

}


// ********************************************************
// VOLUME AND INPUT DISPLAY
// Basic mode where volume and input are shown
//
static void oledUpdateVolIn(void) {

	const unsigned char Y_VOL = 59;
	const unsigned char X_VOL = 0;	
	char buf[OLED_ROW_BUF];
	
	
	// If mute is on, print mute text
	if(preamp.mute == true) {
			
		u8g2_SetFont(&disp, u8g2_font_fub20_tr);
		snprintf_P(buf, 5, PSTR("%S"), PSTR("MUTE"));
		u8g2_DrawStr(&disp, 4, 50, buf);
	}
	else {  // Normal display with volume and input
			
		// Print volume value with large font
		u8g2_SetFont(&disp, u8g2_font_logisoso42_tn);

		snprintf_P(buf, 5, PSTR("%3d"), preamp.masterVol);
		u8g2_DrawStr(&disp, X_VOL, Y_VOL, buf);
			
		// Print "dB" with smaller font
		u8g2_SetFont(&disp, u8g2_font_fub20_tr);
		snprintf_P(buf, 3, PSTR("%S"), PSTR("dB"));
		u8g2_DrawStr(&disp, 92, Y_VOL, buf);
		
		// Show fs for a short period of time, inputFs is used as a flag for showing this value as well
		if(preamp.inputFs != 0) {
			u8g2_SetFont(&disp, u8g2_font_10x20_mf);
			snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%3i kHz"), preamp.inputFs);
			u8g2_DrawStr(&disp, 2, 14, buf);
			
			// Clear the value is it is used as a flag, not very handy but...
			preamp.inputFs = 0;
			resetEvent(DISP_REFRESH_EV, true, false);
			
			u8g2_SetFont(&disp, u8g2_font_fub20_tr); // this is for the next lines
		}
	}
		
	// Input symbol is blinking if controlled to do so (signal unlock)
	if(sys.dispBlink == false) {
		// Print input number inside an empty box
		snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%2s"), preamp.inputName[preamp.input]);
		u8g2_DrawStr(&disp, 88, 24, buf);
		u8g2_DrawFrame(&disp, 86, 0, 39, 28);	
	}

}


// ********************************************************
// MENU DISPLAY
// Shows three menu items at a time when in menu, the selected one highlighted. Keeps the selection
// in the middle.
//
static void oledUpdateMenu(void) {

	const unsigned char Y_MENU[] = {20, 40, 60};
	const unsigned char X_MENU = 5;	
	char buf[OLED_ROW_BUF];
	signed char i;
	
  
	// Font for menu
	u8g2_SetFont(&disp, u8g2_font_10x20_mf);

	// sys.menuIndex starts at 1 but array where strings are is 0-based
	if(sys.menuIndex == 0) {
		i = 1;
	}
	else if(sys.menuIndex == sys.noMenuItems-1) {
		i = sys.noMenuItems-2;
	}
	else {
		i = sys.menuIndex;
	}

	// Shows three items at a time. Selected one (normally in the middle) is highlighted with
	// inverted colors. NOTE: I have noticed that inverting draw colors makes this function 
	// extremely slow causing visible lag. However, I have not tried to modify the u8g2 functions,
	// it is not that bad. Clearly the best option would be to use higher clock speed but this
	// board does not have external crystal for mcu.
	// Item 1
	if(sys.menuIndex == 0) {
		u8g2_SetDrawColor(&disp, 0);
	}
	snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%S"), sys.menuItem[i-1].text);
	u8g2_DrawStr(&disp, X_MENU, Y_MENU[0], buf);
	u8g2_SetDrawColor(&disp, 1);

	// Item 2
	if(sys.menuIndex > 0 && sys.menuIndex < sys.noMenuItems-1) {
		u8g2_SetDrawColor(&disp, 0);
	}
	snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%S"), sys.menuItem[i].text);
	u8g2_DrawStr(&disp, X_MENU, Y_MENU[1], buf);
	u8g2_SetDrawColor(&disp, 1);

	// Item 3
	if(sys.menuIndex == sys.noMenuItems-1) {
		u8g2_SetDrawColor(&disp, 0);
	}
	snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%S"), sys.menuItem[i+1].text);
	u8g2_DrawStr(&disp, X_MENU, Y_MENU[2], buf);
	u8g2_SetDrawColor(&disp, 1);	
}


// ********************************************************
// MENU ITEM SELECTED DISPLAY
// When in menu and menu item is selected, shows the item name and blinks the value.
//
static void oledUpdateMenuItem(void) {

	const unsigned char Y_LOC[] = {20, 50};
	char buf[OLED_ROW_BUF];	
	const char* menuStringPtr;
	unsigned char x_loc;
  
	// Choose font
	u8g2_SetFont(&disp, u8g2_font_10x20_mf);
	
	// Draw name of the selected menu item
	snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%S"), sys.menuItem[sys.menuIndex].text);
	u8g2_DrawStr(&disp, 10, Y_LOC[0], buf);


	// Show and blink the value and unit
	if(sys.dispBlink == false) {
		
		// There are four possibilities for data format: numerical data, numerical data + unit, mapped string, 
		// and mapped string + unit. These are noted by possible NULL in dataMap or unit and must be checked beforehand.
		
		// If dataMap is NULL, numerical data is shown
		if(sys.menuItem[sys.menuIndex].dataMap == NULL) {
			
			// numerical data only
			if(sys.menuItem[sys.menuIndex].unit == NULL) {
				snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%i"), sys.menuItem[sys.menuIndex].data);
			}
			else {	// numerical data + unit
				snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%i %S"), sys.menuItem[sys.menuIndex].data, sys.menuItem[sys.menuIndex].unit);
			}
		}
		else { // Otherwise data is used for indexing menu option strings stored in flash
			
			// dataMap is a pointer to an array of pointers to menu option strings. Therefore, 
			// dataMap + data (positive integer) gives the address of a pointer to the corresponding
			// menu option string. This array of pointers is in PROGMEM, hence pgm_read_word. Not simple but elegant.
			menuStringPtr = (const char*)pgm_read_word(sys.menuItem[sys.menuIndex].dataMap + sys.menuItem[sys.menuIndex].data);
			
			// mapped string only
			if(sys.menuItem[sys.menuIndex].unit == NULL) {
				snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%S"), menuStringPtr);
			}
			else {	// mapped string + unit
				snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%S %S"), menuStringPtr, sys.menuItem[sys.menuIndex].unit);
			}
		}
		// Draw buf
		// X-axis location for data+unit string
		x_loc = pgm_read_byte(sys.menuItem[sys.menuIndex].xLoc);
		u8g2_DrawStr(&disp, x_loc, Y_LOC[1], buf);
	}
}


// ********************************************************
// ERROR DISPLAY
// Indicates error and shows the error code.
//
static void oledUpdateError(void) {
	
	char buf[OLED_ROW_BUF];	
	
	
	// Print "ERROR" and blink error code
	u8g2_SetFont(&disp, u8g2_font_fub20_tr);
	
	snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%S"), PSTR("ERROR"));
	u8g2_DrawStr(&disp, 25, 28, buf);
	
	if(sys.dispBlink == false) {
		snprintf_P(buf, OLED_ROW_BUF+1, PSTR("%S%02x"), PSTR("0x"), sys.errorCode);
		u8g2_DrawStr(&disp, 35, 52, buf);
	}		
			
}


// ********************************************************
// U8G2 FUNCTIONS
// These are needed for u8g2 to function. See u8g2 documentation for more information.
// ********************************************************
// 
// SOFTWARE I2C GPIO AND DELAY IMPLEMENTATION
// This was used in initial tests. The delays are somewhat nonsense and this is extremely slow.
// It is just for debugging purposes.
//
uint8_t u8g2_gpio_delay_sw_i2c(u8x8_t *u8x8, uint8_t msg, uint8_t arg_int, void *arg_ptr)
{
	switch(msg)
	{
		case U8X8_MSG_GPIO_AND_DELAY_INIT:	// called once during init phase of u8g2/u8x8
			break;							// can be used to setup pins
		case U8X8_MSG_DELAY_NANO:			// delay arg_int * 1 nano second
			break;
		case U8X8_MSG_DELAY_100NANO:		// delay arg_int * 100 nano seconds
			break;
		case U8X8_MSG_DELAY_10MICRO:		// delay arg_int * 10 micro seconds
			_delay_us(arg_int*10);
			break;
		case U8X8_MSG_DELAY_MILLI:			// delay arg_int * 1 milli second
			_delay_ms(arg_int*1);
			break;
		case U8X8_MSG_DELAY_I2C:				// arg_int is the I2C speed in 100KHz, e.g. 4 = 400 KHz
			_delay_us(4);
			break;							// arg_int=1: delay by 5us, arg_int = 4: delay by 1.25us
		case U8X8_MSG_GPIO_I2C_CLOCK:		// arg_int=0: Output low at I2C clock pin
			if(arg_int == 1)
        PORTC |= 0x01;
			else
        PORTC &= ~0x01;
			break;							// arg_int=1: Input dir with pullup high for I2C clock pin
		case U8X8_MSG_GPIO_I2C_DATA:			// arg_int=0: Output low at I2C data pin
			if(arg_int == 1)
        PORTC |= 0x02;
			else
        PORTC &= ~0x02;
			break;							// arg_int=1: Input dir with pullup high for I2C data pin
		default:
			u8x8_SetGPIOResult(u8x8, 1);			// default return value
			break;
	}
	return 1;
}


// DUMMY GPIO AND DELAY IMPLEMENTATION FOR HARDWARE I2C
// These are not needed with hardware I2C but the function body needs to exist.
//
uint8_t u8g2_gpio_delay_dummy(u8x8_t *u8x8, uint8_t msg, uint8_t arg_int, void *arg_ptr)
{
	switch(msg)
	{
		case U8X8_MSG_GPIO_AND_DELAY_INIT:
			break;
		case U8X8_MSG_DELAY_NANO:
			break;
		case U8X8_MSG_DELAY_100NANO:
			break;
		case U8X8_MSG_DELAY_10MICRO:
			break;
		case U8X8_MSG_DELAY_MILLI:
			break;
		case U8X8_MSG_DELAY_I2C:
			break;
		case U8X8_MSG_GPIO_I2C_CLOCK:
			break;
		case U8X8_MSG_GPIO_I2C_DATA:
			break;
		default:
			u8x8_SetGPIOResult(u8x8, 1);			// default return value
			break;
	}
	return 1;
}


// HARDWARE I2C LOW-LEVEL IMPLEMENTATION
// This handles sending bytes using hardware I2C block. All pins and registers are hard-coded
// but easy to change if copied elsewhere. Switch case body is from u8g2 documentation. Note 
// that this is in no way clever implementation, for example there are no timeouts for while
// loops. But it seems to work.
//
uint8_t u8g2_avr_hw_i2c(u8x8_t *u8x8, uint8_t msg, uint8_t arg_int, void *arg_ptr)
{
	uint8_t *data;
	
	switch(msg)
	{
		case U8X8_MSG_BYTE_SEND:
			data = (uint8_t *)arg_ptr;
		
			while( arg_int > 0 ) {
				TWDR = *data;
				TWCR = (1<<TWINT) | (1<<TWEN);
				data++;
				arg_int--;
				while (!(TWCR & (1<<TWINT)));	// confirm transmission - display may not sent ACK
				// should check some status here?
			}
			break;
		
		case U8X8_MSG_BYTE_INIT:
			//TWBR =			// bit rate
			TWSR = 0x02;		// prescaler 16
			DDRC &= ~0x03;		// set pins as inputs wo pull-up (there are external resistors)
			PORTC  &= ~0x03;
			break;
		
			case U8X8_MSG_BYTE_SET_DC:
			break;
		
		case U8X8_MSG_BYTE_START_TRANSFER:

			TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);		// start condition
			while (!(TWCR & (1<<TWINT)));					// wait for flag to confirm start sent
			// should check some status here?
		
			TWDR = u8x8_GetI2CAddress(u8x8);				// send the address and write bit
			TWCR = (1<<TWINT) | (1<<TWEN);
			while (!(TWCR & (1<<TWINT)));					// ensure proper transmission - ACK note!
			// should check some status here?
		
			break;
		
		case U8X8_MSG_BYTE_END_TRANSFER:
			//i2c_stop(u8x8);
			TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);	// stop condition
			break;
		
		default:
			return 0;
	}
	return 1;
}

